"""Connect to another Kukur instance using Arrow Flight."""
# SPDX-FileCopyrightText: 2021 Timeseer.AI
#
# SPDX-License-Identifier: Apache-2.0
from .kukur import from_config

__all__ = ["from_config"]
