"""Dummy source for integration tests."""

# SPDX-FileCopyrightText: 2022 Timeseer.AI
# SPDX-License-Identifier: Apache-2.0

from .integration_test_source import from_config

__all__ = ["from_config"]
